#
#  Copyright (c) 2008 INdT - Instituto Nokia de Tecnologia
#
#  This file is part of carman-python.
#
#  carman-python is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  carman-python is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from common.carmanconfig import CarmanConfig, METRIC
from main.messagedlgctrl import MessageDlgController
from maps.infosharingview import BuddyView, BuddyInformationView
from models.buddymodel import BuddyModel
from models.gpsmodel import GPSModel
from models.mapmodel import MapModel
from models.infosharingmodel import InfoSharingModel


class InfoSharingController(object):

    def __init__(self, controller, canvas):
        self.__move_cb = None

        self.zoom = None
        self.canvas = canvas

        self.buddy_model = BuddyModel()
        self.buddy_model.add_buddy_connect_cb(self.__buddy_connect_cb)
        self.buddy_model.add_buddy_disconnect_cb(self.__buddy_disconnect_cb)
        self.buddy_model.add_request_created_cb(self.__buddy_connect_cb)
        self.buddy_model.add_request_removed_cb(self.__buddy_disconnect_cb)

        self.map_model = MapModel()

        self.is_model = InfoSharingModel()

        metric = CarmanConfig().get_unit() == METRIC
        self.buddy_view = BuddyView(controller, canvas)
        self.buddy_information = BuddyInformationView(canvas, metric)

    def __buddy_data_available(self, name):
        if name == self.buddy_model.active_buddy:
            self.update_buddy_information()

    def __buddy_connect_cb(self, name):
        if name == self.buddy_model.active_buddy:
            alias = self.is_model.get_buddy_alias(name)
            if alias:
                self.buddy_information.set_buddy_name(alias)
            else:
                self.buddy_information.set_buddy_name(name)

            if self.buddy_model.is_buddy_connected(name):
                self.buddy_model.add_data_available_cb(self.__buddy_data_available)
                if not self.buddy_information.is_visible():
                    self.buddy_information.set_speed_value(-1)
                    self.buddy_information.set_rpm_value(-1)
                self.buddy_information.show_data_values()
            else:
                self.buddy_information.show_buddy_connecting()
            self.buddy_information.show()

    def __buddy_disconnect_cb(self, name, reason=None, connected = False):
        alias = self.is_model.get_buddy_alias(name)
        if not alias:
            alias = name

        if self.buddy_model.has_buddies():
            self.update_buddy_information()
        else:
            if connected:
                self.buddy_model.del_data_available_cb(self.__buddy_data_available)
            self.buddy_view.deactivate()
            self.buddy_information.hide()
            self.set_move_cb(None)

        if connected:
            msg = MessageDlgController()
            if reason:
                msg.show_message("%s<br>disconnected:<br>%s" % \
                    (alias, reason), title="BUDDY DISCONNECTED")
            else:
                msg.show_message("%s<br>disconnected" % \
                    (alias), title="BUDDY DISCONNECTED")

    def update_buddy_information(self):
        def update_buddy_view(buddy_gps):
            self.buddy_view.set_image(buddy_gps)

            x, y = [item >> self.zoom for item in \
                        self.map_model.latlon_to_xy(lat, lon)]
            self.buddy_view.activate()
            self.buddy_view.set_buddy_position(x, y)
            if callable(self.__move_cb):
                self.__move_cb(x, y)

        active_buddy = self.buddy_model.active_buddy
        alias = self.is_model.get_buddy_alias(active_buddy)
        if alias:
            self.buddy_information.set_buddy_name(alias)
        else:
            self.buddy_information.set_buddy_name(active_buddy)

        if self.buddy_model.is_buddy_connected(active_buddy):
            (lat, lon, alt, speed, track, rpm, gps_status, obd_status) = \
                self.buddy_model.get_connected_buddy_data(active_buddy)

            self.buddy_information.set_speed_value(speed)
            self.buddy_information.set_rpm_value(rpm)
            self.buddy_information.set_gps_status(gps_status)
            self.buddy_information.show_data_values()

            if self.zoom != None:
                if gps_status == GPSModel.FIXED:
                    update_buddy_view(True)
                elif gps_status != GPSModel.DISABLED and \
                        self.buddy_model.buddy_has_view():
                    update_buddy_view(False)
                else:
                    self.buddy_view.deactivate()
        else:
            self.buddy_information.show_buddy_connecting()
            self.buddy_view.deactivate()

    def set_move_cb(self, cb):
        if callable(cb):
            self.__move_cb = cb
            if self.buddy_model.active_buddy != '':
                x, y = [item >> self.zoom for item in \
                    self.map_model.latlon_to_xy( \
                        *self.buddy_model.get_connected_buddy_data()[0:2])]
                self.__move_cb(x, y, False)
        else:
            self.__move_cb = None

    def get_lock(self):
        return callable(self.__move_cb)

    def set_unit_system(self, value):
        self.buddy_information.set_unit_system(value)

    def set_view(self, theme, view):
        if self.buddy_model.is_buddy_connected( \
                        self.buddy_model.active_buddy):
            gps_status = self.buddy_model.get_connected_buddy_data( \
                                self.buddy_model.active_buddy)[6]
            if gps_status == GPSModel.FIXED:
                self.buddy_view.set_theme(theme, True)
            elif gps_status != GPSModel.DISABLED and \
                    self.buddy_model.buddy_has_view():
                self.buddy_view.set_theme(theme, False)
            else:
                self.buddy_view.set_theme(theme)
        else:
            self.buddy_view.set_theme(theme)
        self.buddy_view.set_buddy_area(view.part_geometry_get("area_view"))
        view.part_swallow("buddy_view", self.buddy_view)
        self.buddy_information.set_view(theme, view)
        self.update_buddy_information()

    def set_fullscreen(self, fullscreen):
        self.buddy_view.set_fullscreen(fullscreen)

    def get_buddy_position(self):
        return self.buddy_view.get_buddy_position()

    def set_position(self, x, y, zoom):
        self.buddy_view.set_position(x, y)
        if self.zoom != zoom:
            self.zoom = zoom
            if self.buddy_model.active_buddy != '':
                x, y = [item >> self.zoom for item in \
                    self.map_model.latlon_to_xy( \
                        *self.buddy_model.get_connected_buddy_data()[0:2])]
                self.buddy_view.set_buddy_position(x, y)

    def finalize(self):
        self.buddy_model.disconnect_all_buddies()
