#
#  Copyright (c) 2008 INdT - Instituto Nokia de Tecnologia
#
#  This file is part of carman-python.
#
#  carman-python is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  carman-python is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import edje
from common.carmanconfig import CarmanConfig
from main.container import Controller
from main.mainview import MainView
from models.gpsmodel import GPSModel
from models.kbmodel import KeyBoardModel
from models.mapmodel import MapModel
from models.buddymodel import BuddyModel
from models.connmodel import ConnectionModel
from models.infosharingmodel import InfoSharingModel
from maps.optionctrl import OptionController
from maps.maptilectrl import MapTileController
from maps.progressbarctrl import ProgressBarController
from maps.trackctrl import TrackController, TrackAuxController
from maps.infosharingctrl import InfoSharingController
from maps.connectaccountctrl import ConnectAccountController
from maps.connectbuddyctrl import ConnectBuddyController
from main.messagedlgctrl import MessageDlgController


class MapsController(Controller):

    def __init__(self, parent, canvas, theme):
        self.canvas = canvas
        self._parent = parent
        self.screen = None
        self.download_ctrl = None
        self.__toggle_fullscreen = False
        self.buddy_model = BuddyModel()
        self.buddy_model.add_buddy_connect_cb(self.goto_buddy_cb)
        self.buddy_model.add_buddy_disconnect_cb(self.goto_buddy_cb)
        self.buddy_model.add_request_created_cb(self.goto_buddy_cb)
        self.buddy_model.add_request_removed_cb(self.goto_buddy_cb)

        self.main_view = MainView()
        self.map_model = MapModel()
        self.map_model.add_map_service_updated(self.check_map_service_zoom)

        self.gps_model = GPSModel()
        self.gps_model.add_status_changed_cb(self.set_gps_status)

        self.is_model = InfoSharingModel()
        if CarmanConfig().is_service:
            self.is_model.add_status_connect_cb(self.set_is_status)
            self.is_model.add_network_error_cb(self.reconnect_account)
        self.conn_model = ConnectionModel()
        self.conn_model.add_connection_lost_cb(self.connection_lost)

        self.track_ctrl = TrackController(canvas, self.gps_model)
        self.track_aux_ctrl = TrackAuxController(canvas)
        self.is_ctrl = InfoSharingController(self, canvas)
        self.map_ctrl = MapTileController(self, canvas, self.track_ctrl,
            self.is_ctrl, self.track_aux_ctrl)
        self.map_ctrl.set_zoom_changed_cb(self.update_zoom_buttons)
        self.progressbar_ctrl = ProgressBarController(canvas)

        self.set_theme(theme)
        lat, lon, self.zoom = CarmanConfig().get_last_fix()
        self.map_ctrl.set_position(lat, lon, self.zoom)

        self.kb_model = KeyBoardModel()
        self.option = OptionController(self)

    def __on_buddy_pressed(self, *data):
        if CarmanConfig().is_service:
            if not self.is_model.service:
                self.is_model.instantiate_isservice()

            if self.is_model.account_status == InfoSharingModel.CONNECTED:
                if self.buddy_model.size > 1:
                    self.buddy_model.set_next_active_buddy()
                    self.is_ctrl.update_buddy_information()
                    if self.buddy_model.buddy_has_view():
                        self.set_center_buddy()
                    else:
                        self.set_center_gps()
                else:
                    buddy = ConnectBuddyController(self)
                    if buddy.load_list():
                        buddy.show()
            else:
                account = ConnectAccountController()
                if account.exist_account():
                    account.connect_account()
        else:
            msg = MessageDlgController()
            msg.show_message("Infoshare not installed", title="INFOSHARE")

    def __on_zoom_up(self, *params):
        self.map_ctrl.zoom_up()

    def __on_zoom_down(self, *params):
        self.map_ctrl.zoom_down()

    def check_map_service_zoom(self, repo):
        zoom = self.map_ctrl.zoom
        if zoom is None:
            zoom = self.zoom
        zoom_range = repo.get_zoom()
        if zoom < zoom_range[0]:
            for x in xrange(zoom, zoom_range[0]):
                self.map_ctrl.zoom_down()

        if zoom > zoom_range[-1]:
            for x in xrange(zoom_range[-1], zoom):
                self.map_ctrl.zoom_up()

    def get_view(self):
        return self.screen

    def get_icon(self):
        return (self.theme_file, "images/icon-maps")

    def show_options(self):
        self.option.show()

    def activate(self):
        self.screen.show()
        self.map_ctrl.activate()
        self.kb_model.add_key_down_cb(self.key_down_cb)

    def deactivate(self):
        self.kb_model.del_key_down_cb(self.key_down_cb)

    def key_down_cb(self, obj, event):
        if self.main_view.get_transiting():
            return False

        if event.key == "Left":
            self.map_ctrl.move(-32, 0)
        elif event.key == "Right":
            self.map_ctrl.move(32, 0)
        elif event.key == "Up":
            self.map_ctrl.move(0, -32)
        elif event.key == "Down":
            self.map_ctrl.move(0, 32)
        elif event.key == "F4":
            self.show_options()
        elif event.key == "F6":
            self.fullscreen()
        elif event.key == "F7":
            self.__on_zoom_up()
        elif event.key == "F8":
            self.__on_zoom_down()
        else:
            return False

        return True

    def set_global_unit(self, system):
        self._parent.set_unit_system(system)

    def set_unit_system(self, value):
        self.is_ctrl.set_unit_system(value)

    def set_global_theme(self):
        self._parent.set_theme()

    def set_theme(self, theme):
        if self.screen:
            self.screen.delete()
        self.theme_file = theme
        first_time = self.screen is None
        self.screen = edje.Edje(self.canvas, file=self.theme_file,
                                group="maps")
        self.screen.size = self.canvas.size

        self.track_ctrl.set_view(theme, self.screen)
        self.track_aux_ctrl.set_view(theme, self.screen)
        self.is_ctrl.set_view(theme, self.screen)
        self.map_ctrl.set_view(theme, self.screen)
        self.progressbar_ctrl.set_view(theme, self.screen)

        if not first_time:
            self.update_zoom_buttons(*self.map_ctrl.get_zoom_state())
            self.map_ctrl.update_position()
        self.set_gps_status(self.gps_model.Status())

        if CarmanConfig().is_service:
            self.set_is_status(self.is_model.account_status)

        self.screen.signal_callback_add("zoom-up-pressed", "",
                                        self.__on_zoom_up)
        self.screen.signal_callback_add("zoom-down-pressed", "",
                                        self.__on_zoom_down)
        self.screen.signal_callback_add("goto-gps-pressed", "",
                                        self.set_center_gps)
        self.screen.signal_callback_add("goto-buddy-pressed", "",
                                        self.__on_buddy_pressed)

    def set_gps_status(self, status):
        if status == GPSModel.FIXED:
            self.screen.signal_emit("enable-goto-gps", "")
        else:
            self.screen.signal_emit("disable-goto-gps", "")

    def set_is_status(self, status):
        if status == InfoSharingModel.CONNECTED:
            if self.buddy_model.size > 1:
                self.screen.signal_emit("enable-next-buddy", "")
            else:
                self.screen.signal_emit("enable-goto-buddy", "")
        elif status == InfoSharingModel.CONNECTING:
            self.screen.signal_emit("init-acc-anime", "")
        elif status == InfoSharingModel.DISCONNECTED:
            self.screen.signal_emit("disable-goto-buddy", "")

    def goto_buddy_cb(self, name, reason=None, connected=None):
        if self.buddy_model.size <= 1:
            self.screen.signal_emit("enable-goto-buddy", "")
        else:
            self.screen.signal_emit("enable-next-buddy", "")

    def finalize(self):
        self.is_ctrl.finalize()
        self.track_ctrl.finalize()
        self.track_aux_ctrl.finalize()
        lat, lon, zoom = self.map_ctrl.get_position()
        CarmanConfig().set_last_fix(lat, lon, zoom)
        self.map_ctrl.finalize()

    def update_zoom_buttons(self, up, down):
        if up:
            self.screen.signal_emit("enable-zoom-up", "")
        else:
            self.screen.signal_emit("disable-zoom-up", "")
        if down:
            self.screen.signal_emit("enable-zoom-down", "")
        else:
            self.screen.signal_emit("disable-zoom-down", "")

    def fullscreen(self):
        controller = self.main_view.controller
        view = self.main_view.view

        if self.__toggle_fullscreen:
            view.signal_callback_add("exit-button-pressed", "*",
                                     controller.exit_button)
            view.signal_callback_add("change-button-pressed", "*",
                                     controller.change_button)
            view.signal_callback_add("option-button-pressed", "*",
                                     controller.option_button)
            view.signal_callback_add("status-button-pressed", "*",
                                     controller.status_button)
            self.screen.signal_callback_add("zoom-up-pressed", "",
                                            self.__on_zoom_up)
            self.screen.signal_callback_add("zoom-down-pressed", "",
                                            self.__on_zoom_down)
            self.screen.signal_callback_add("goto-gps-pressed", "",
                                            self.set_center_gps)
            self.screen.signal_callback_add("goto-buddy-pressed", "",
                                            self.__on_buddy_pressed)
            view.signal_emit("default", "")
            self.screen.signal_emit("default", "")
        else:
            view.signal_callback_del("exit-button-pressed", "*",
                                     controller.exit_button)
            view.signal_callback_del("change-button-pressed", "*",
                                     controller.change_button)
            view.signal_callback_del("option-button-pressed", "*",
                                     controller.option_button)
            view.signal_callback_del("status-button-pressed", "*",
                                     controller.status_button)
            self.screen.signal_callback_del("zoom-up-pressed", "",
                                            self.__on_zoom_up)
            self.screen.signal_callback_del("zoom-down-pressed", "",
                                            self.__on_zoom_down)
            self.screen.signal_callback_del("goto-gps-pressed", "",
                                            self.set_center_gps)
            self.screen.signal_callback_del("goto-buddy-pressed", "",
                                            self.__on_buddy_pressed)
            view.signal_emit("fullscreen", "")
            self.screen.signal_emit("fullscreen", "")
        self.__toggle_fullscreen = not self.__toggle_fullscreen
        self.is_ctrl.set_fullscreen(self.__toggle_fullscreen)

    def set_center_buddy(self):
        self.track_ctrl.set_move_cb(None)
        self.is_ctrl.set_move_cb(self.map_ctrl.set_position_xy)
        self.is_ctrl.update_buddy_information()

    def set_center_free(self):
        self.track_ctrl.set_move_cb(None)
        self.is_ctrl.set_move_cb(None)

    def set_center_gps(self, *data):
        if self.gps_model.Status() == GPSModel.FIXED:
            self.is_ctrl.set_move_cb(None)
            self.track_ctrl.set_move_cb(self.map_ctrl.set_position_xy)

    def set_track_aux(self, models, track_file):
        self.track_aux_ctrl.set_track_model(models, track_file)
        # FIXME: Point screen to loaded track
        self.map_ctrl.update_position()

    def get_selected_track(self):
        return self.track_aux_ctrl.get_selected_track()

    def unload_track_aux(self):
        self.track_aux_ctrl.clear()

    def get_map_view_area(self):
        return self.map_ctrl.get_view_area()

    def show_progressbar(self, ctrl):
        self.download_ctrl = ctrl
        self.progressbar_ctrl.show()

    def hide_progressbar(self):
        self.download_ctrl = None
        self.progressbar_ctrl.hide()

    def update_progressbar(self, pos):
        self.progressbar_ctrl.update(pos)

    def is_downloading(self):
        return self.download_ctrl is not None

    def stop_download(self):
        if self.download_ctrl:
            self.download_ctrl.stop_download()
            self.hide_progressbar()

    def reconnect_account(self):
        if ConnectionModel().Status() == ConnectionModel.CONNECTED:
            reconnect = ConnectAccountController()
            if reconnect.exist_account():
                reconnect.connect_account()
        else:
            msg = MessageDlgController()
            msg.show_message("Wlan Disconnected", title="CONNECTION LOST")

    def connection_lost(self):
        if CarmanConfig().is_service:
            account = self.is_model.service.get_account()
            self.is_model.service.account_disconnect()
            self.conn_model.add_connection_enabled_cb(self.reconnect_account)
