#
#  Copyright (c) 2008 INdT - Instituto Nokia de Tecnologia
#
#  This file is part of carman-python.
#
#  carman-python is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  carman-python is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import ecore

from common.singleton import Singleton
from models.gpsmodel import GPSModel
from models.obdmodel import OBDModel
from main.messagedlgctrl import MessageDlgController

class BuddyModel(Singleton):

    # Default values
    (BUDDY_CONNECTING, BUDDY_CONNECTED) = range(2)
    MAX_BUDDIES = 4

    BUDDY_REQUEST_TIMEOUT = 55.0
    USER_REQUEST_TIMEOUT = 60.0

    def __init__(self):
        Singleton.__init__(self)

        self.__buddies = {}
        self.__active_buddy = ''

        self.__buddy_data = {}
        self.__has_view = []
        self.__buddy_connect_cbs = []
        self.__buddy_disconnect_cbs = []
        self.__data_available_cbs = []

        self.__requests = {}
        self.__request_created_cbs = []
        self.__request_removed_cbs = []

    def __get_active_buddy(self):
        return self.__active_buddy
    active_buddy = property(__get_active_buddy)

    def __get_requests(self):
        return self.__requests.keys()
    requests = property(__get_requests)

    def __get_size(self):
        return len(self.__buddies)
    size = property(__get_size)

    def __request_timeout(self, name, msg=None):
        if self.has_request(name):
            self.remove_request(name)

            if msg:
                msg.cancel_cb()
                msg = MessageDlgController()
                msg.show_message("Invitation from <br>%s<br>has expired" % \
                        name, title="BUDDY REQUEST")
            else:
                msg = MessageDlgController()
                msg.show_message("Request timeout for <br>%s" % \
                        name, title="BUDDY REQUEST")

    def add_buddy_connect_cb(self, cb):
        if callable(cb) and cb not in self.__buddy_connect_cbs:
            self.__buddy_connect_cbs.append(cb)

    def del_buddy_connect_cb(self, cb):
        if cb in self.__buddy_connect_cbs:
            self.__buddy_connect_cbs.remove(cb)

    def add_buddy_disconnect_cb(self, cb):
        if callable(cb) and cb not in self.__buddy_disconnect_cbs:
            self.__buddy_disconnect_cbs.append(cb)

    def del_buddy_disconnect_cb(self, cb):
        if cb in self.__buddy_disconnect_cbs:
            self.__buddy_disconnect_cbs.remove(cb)

    def add_data_available_cb(self, cb):
        if callable(cb) and cb not in self.__data_available_cbs:
            self.__data_available_cbs.append(cb)

    def del_data_available_cb(self, cb):
        if cb in self.__data_available_cbs:
            self.__data_available_cbs.remove(cb)

    def add_request_created_cb(self, cb):
        if callable(cb) and cb not in self.__request_created_cbs:
            self.__request_created_cbs.append(cb)

    def del_request_created_cb(self, cb):
        if cb in self.__request_created_cbs:
            self.__request_created_cbs.remove(cb)

    def add_request_removed_cb(self, cb):
        if callable(cb) and cb not in self.__request_removed_cbs:
            self.__request_removed_cbs.append(cb)

    def add_received_request_timeout(self, name, alias, accepted_cb, rejected_cb):
        def confirm_cb():
            self.remove_request(name)
            if callable(accepted_cb):
                accepted_cb()

        def cancel_cb():
            self.remove_request(name)
            if callable(rejected_cb):
                rejected_cb()

        msg = MessageDlgController(confirm_cb=confirm_cb, cancel_cb=cancel_cb)

        if alias != None:
            msg.show_message("%s<br>wants to share Carman Information." \
                    "<br>Do you accept?" % alias, buttons=2)
        else:
            msg.show_message("%s<br>wants to share Carman Information." \
                    "<br>Do you accept?" % name, buttons=2)

        if not self.has_buddies():
            self.__active_buddy = name
        self.__requests[name] = ecore.Timer(self.BUDDY_REQUEST_TIMEOUT, \
                self.__request_timeout, name, msg)
        self.__buddies[name] = self.BUDDY_CONNECTING

    def add_sent_request_timeout(self, name):
        if not self.has_buddies():
            self.__active_buddy = name
        self.__requests[name] = ecore.Timer(self.USER_REQUEST_TIMEOUT, \
                self.__request_timeout, name)
        self.__buddies[name] = self.BUDDY_CONNECTING

        for cb in self.__request_created_cbs:
            cb(name)

    def buddy_has_view(self, name=None):
        if name == None:
            name = self.__active_buddy
        return name in self.__has_view

    def received_data_msg(self, name, message):
        if not self.is_buddy_connected(name):
            # Ignore since buddy is not properly connected
            return True

        gps_status = message.split(";")[8]
        obd_status = message.split(";")[9]

        new_data = [float(item) for item in message.split(";")[2:8]]
        new_data.append(gps_status)
        new_data.append(obd_status)

        if gps_status == GPSModel.DISABLED:
            new_data[0] = float(-1) # lat
            new_data[1] = float(-1) # lon
            new_data[2] = float(-1) # alt
            new_data[4] = float(-1) # track

        if obd_status != OBDModel.CONNECTED:
            new_data[5] = float(-1) # rpm
            if gps_status != GPSModel.FIXED:
                new_data[3] = float(-1) # speed

        self.__buddy_data[name] = new_data

        if name in self.__has_view:
            if gps_status == GPSModel.DISABLED:
                self.__has_view.remove(name)
        elif gps_status == GPSModel.FIXED:
            self.__has_view.append(name)

        for cb in self.__data_available_cbs:
            cb(name)
        return True

    def received_close_conn_msg(self, name):
        if self.buddy_exists(name):
            self.disconnect_buddy(name, reason="Buddy closed connection")
        return True

    def get_connected_buddy_data(self, name=None):
        if name == None:
            name = self.__active_buddy
        if name in self.__buddy_data:
            return self.__buddy_data[name]
        else:
            return None

    def get_connected_buddies(self):
        return [buddy for buddy in self.__buddies if \
            self.is_buddy_connected(buddy)]

    def get_connecting_buddies(self):
        return [buddy for buddy in self.__buddies if \
            self.is_buddy_connecting(buddy)]

    def is_buddy_connected(self, name):
        return self.buddy_exists(name) and \
                self.__buddies[name] == self.BUDDY_CONNECTED

    def is_buddy_connecting(self, name):
        return self.buddy_exists(name) and \
                self.__buddies[name] == self.BUDDY_CONNECTING

    def buddy_exists(self, name):
        return name in self.__buddies

    def has_buddies(self):
        return len(self.__buddies) > 0

    def has_request(self, name):
        return name in self.__requests

    def connect_buddy(self, name):
        if not self.has_buddies():
            self.__active_buddy = name
        self.__buddies[name] = self.BUDDY_CONNECTED
        self.__buddy_data[name] = [float(-1), float(-1), float(-1), \
            float(-1), float(-1), float(-1), \
            GPSModel.DISABLED, OBDModel.DISABLED]
        for cb in self.__buddy_connect_cbs:
            cb(name)

    def disconnect_buddy(self, name, reason=None):
        if self.buddy_exists(name):
            self.remove_buddy(name)
            del self.__buddy_data[name]
            if name in self.__has_view:
                self.__has_view.remove(name)
            for cb in self.__buddy_disconnect_cbs:
                cb(name, reason, connected = True)

    def disconnect_all_buddies(self):
        for buddy in self.get_connected_buddies():
            self.disconnect_buddy(buddy)
        for buddy in self.get_connecting_buddies():
            self.remove_request(request)

    def remove_request(self, name):
        if self.has_request(name):
            self.remove_buddy(name)
            self.__requests[name].delete()
            del self.__requests[name]
            for cb in self.__request_removed_cbs:
                cb(name)

    def remove_buddy(self, name):
        if self.buddy_exists(name):
            del self.__buddies[name]
            if name == self.__active_buddy:
                self.set_next_active_buddy()

    def set_next_active_buddy(self):
        if self.has_buddies():
            i = self.__buddies.keys().index(self.__active_buddy)
            i = (i + 1) % self.size
            self.__active_buddy = self.__buddies.keys()[i]
        else:
            self.__active_buddy = ''
