#
#  Copyright (c) 2008 INdT - Instituto Nokia de Tecnologia
#
#  This file is part of carman-python.
#
#  carman-python is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  carman-python is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import ecore
import re

from common.carmanconfig import CarmanConfig
from common.singleton import Singleton
from main.messagedlgctrl import MessageDlgController
from models.buddymodel import BuddyModel
from models.connmodel import ConnectionModel
from models.dbusmodel import CarmandDbusModel
from models.gpsmodel import GPSModel
from models.obdmodel import OBDModel

GREETING_PATTERN   = "^.*cds;ccs;cde;$"
ACCEPTED_PATTERN   = "^.*cds;cca;cde;$"
REJECTED_PATTERN   = "^.*cds;ccr;cde;$"
CLOSE_CONN_PATTERN = "^.*cds;cce;cde;$"
DATA_PATTERN       = "^.*cds;dds;(([\\-]*[0-9]+\\.[0-9]+;){6})(([A-Z][a-z]+;){2})dde;cde;$"

GREETING_FMT       = "cds;ccs;cde;"
ACCEPTED_FMT       = "cds;cca;cde;"
REJECTED_FMT       = "cds;ccr;cde;"
CLOSE_CONN_FMT     = "cds;cce;cde;"
DATA_FMT           = "cds;dds;%f;%f;%f;%f;%f;%f;%s;%s;dde;cde;"

GREETING_MESSAGE   = "I'm trying to share my Carman information with you, but it appears that you're not using Carman as your IM client. If you don't want to share any information, please ignore this message.<br>For more information about Carman, please visit: http://openbossa.indt.org/carman<br><span style='color: #ffffff'>%s</span>"

NO_IM_MESSAGE      = "You are trying to communicate with me, but I'm currently using Carman and only accepting Carman-like messages. For more information about Carman, please visit: http://openbossa.indt.org/carman<br><span style='color: #ffffff'>%s</span>"

# Default data report timer value
__DATA_REPORT_TIME__ = 3.0


class InfoSharingModel(Singleton):

    # Account states
    DISCONNECTED = "Disconnected"
    CONNECTING   = "Connecting"
    CONNECTED    = "Connected"

    def __init__(self):
        """
        Initializes InfoSharingModel
        """
        Singleton.__init__(self)

        self.__account_status = self.DISCONNECTED
        self.__config = CarmanConfig()
        self.__connection_error = False
        self.__data_timer = None
        self.__is_service = None

        self.__connection_error_cbs = []
        self.__buddy_signed_off_cbs = []
        self.__buddy_signed_on_cbs = []
        self.__network_error_cbs = []
        self.__signed_off_cbs = []
        self.__signed_on_cbs = []
        self.__status_changed_cbs = []

        self._is_isservice = False

        # lat, lon, alt, speed, track, rpm, gps-status, obd-status
        self.data = [float(-1), float(-1), float(-1), float(-1), float(-1), \
            float(-1), GPSModel.DISABLED, OBDModel.DISABLED]

        self.carmand_model = CarmandDbusModel()

        self.buddy_model = BuddyModel()
        self.buddy_model.add_buddy_connect_cb(self.__buddy_connected_cb)
        self.buddy_model.add_buddy_disconnect_cb(self.__buddy_disconnected_cb)

        self.conn_model = ConnectionModel()
        self.conn_model.add_connection_lost_cb(self.__connection_lost_cb)

        self.gps_model = GPSModel()
        self.gps_model.add_data_available_cb(self.__gps_data_available_cb)

        self.obd_model = OBDModel()
        self.obd_model.add_data_available_cb((("0C", 1, 0), ), \
                                             self.__obd_rpm_available_cb)
        self.obd_model.add_data_available_cb((("0D", 1, 0), ), \
                                             self.__obd_speed_available_cb)

        self.add_signed_on_cb(self.__signed_on_cb)
        self.add_signed_off_cb(self.__signed_off_cb)
        self.add_connection_error_cb(self.__connection_error_cb)
        self.add_buddy_signed_off_cb(self.__buddy_signed_off_cb)

    def instantiate_isservice(self):
        from models.isservice import ISService
        self.__is_service = ISService(self)

    def __get_service(self):
        return self.__is_service
    service = property(__get_service)

    def __gps_data_available_cb(self, *data):
        # when GPS is stopped, data items can get a 'nan'
        if all([item == item for item in data]):
            if self.obd_model.Status() != OBDModel.CONNECTED:
                self.data[3] = float(data[3]) # speed (if OBD is not connected)

            if self.data[3] > 0: # car is moving
                self.data[0] = float(data[0]) # lat
                self.data[1] = float(data[1]) # lon
                self.data[2] = float(data[2]) # alt
                self.data[4] = float(data[4]) # track

    def __obd_rpm_available_cb(self, model, pid, *data):
        self.data[5] = float(data[0]) # rpm

    def __obd_speed_available_cb(self, model, pid, *data):
        self.data[3] = float(data[0]) # speed

    def __buddy_connected_cb(self, name):
        if not self.__data_timer:
            self.__data_timer = ecore.Timer(__DATA_REPORT_TIME__, \
                self.send_data_cb)

    def __buddy_disconnected_cb(self, name, reason=None, connected=None):
        self.send_close_connection(name)
        if not self.buddy_model.has_buddies() and self.__data_timer:
            self.__data_timer.delete()
            self.__data_timer = None

    def __connection_error_cb(self, username, protocol_id, short_desc, \
            desc=None):
        self.buddy_model.disconnect_all_buddies()

        short_desc = short_desc.split("\n")[0]
        def reconnect_account_cb():
            disconnect_account_cb()
            for cb in self.__network_error_cbs:
                cb()
            self._set_account_status(self.CONNECTING)

        def disconnect_account_cb():
            self.__is_service.account_disconnect()
            self._set_account_status(self.DISCONNECTED)

        self.__connection_error = not self.__connection_error
        if short_desc == "Network error":
            msg = MessageDlgController(confirm_cb=reconnect_account_cb,
                    cancel_cb=disconnect_account_cb)
            msg.show_message("%s<br>Reconnect account?" % \
                    short_desc, buttons=2, title="ACCOUNT STATUS")
        else:
            msg = MessageDlgController(confirm_cb=disconnect_account_cb)
            if desc:
                msg.show_message("Connection error:<br>%s<br>%s" % \
                        (short_desc, desc), title="ACCOUNT STATUS")
            else:
                msg.show_message("Connection error:<br>%s" % short_desc, \
                        title="ACCOUNT STATUS")

    def __connection_lost_cb(self):
        self.buddy_model.disconnect_all_buddies()

    def __signed_on_cb(self, username, protocol_id):
        protocols = self.service.get_protocols()
        msg = MessageDlgController()
        msg.show_message("You are now signed on<br>%s<br>(%s)" % \
            (username.split("/")[0], protocols[protocol_id][0]), \
            title="ACCOUNT STATUS")
        self._set_account_status(self.CONNECTED)
        self.__config.last_account_status = 'online'

    def __signed_off_cb(self, username, protocol_id):
        if self.__connection_error:
            self.__connection_error = not self.__connection_error
        else:
            self.buddy_model.disconnect_all_buddies()

            protocols = self.service.get_protocols()
            msg = MessageDlgController()
            msg.show_message("You are now signed off<br>%s<br>(%s)" % \
                (username.split("/")[0], protocols[protocol_id][0]), \
                title="ACCOUNT STATUS")
            self._set_account_status(self.DISCONNECTED)
        self.__config.last_account_status = 'offline'

    def __buddy_signed_off_cb(self, name, alias):
        if self.buddy_model.is_buddy_connected(name):
            self.buddy_model.disconnect_buddy(name, reason="Connection lost")

    def _on_signed_on_cb(self, username, protocol_id):
        for cb in self.__signed_on_cbs:
            cb(username, protocol_id)

    def _on_signed_off_cb(self, username, protocol_id):
        for cb in self.__signed_off_cbs:
            cb(username, protocol_id)

    def _on_connection_error_cb(self, username, protocol_id, short_desc, desc=None):
        for cb in self.__connection_error_cbs:
            if desc:
                cb(username, protocol_id, short_desc, desc)
            else:
                cb(username, protocol_id, short_desc)

    def add_status_connect_cb(self, cb):
        if callable(cb) and cb not in self.__status_changed_cbs:
            self.__status_changed_cbs.append(cb)

    def del_status_connect_cb(self, cb):
        if cb in self.__status_changed_cbs:
            self.__status_changed_cbs.remove(cb)

    def _set_account_status(self, status):
        if status in [self.CONNECTED, self.CONNECTING, self.DISCONNECTED]:
            for cb in self.__status_changed_cbs:
                cb(status)
            self.__account_status = status

    def _get_account_status(self):
        return self.__account_status

    account_status = property(_get_account_status, _set_account_status)

    def _on_buddy_signed_on_cb(self, name, alias):
        for cb in self.__buddy_signed_on_cbs:
            cb(name, alias)

    def _on_buddy_signed_off_cb(self, name, alias):
        for cb in self.__buddy_signed_off_cbs:
            cb(name, alias)

    def _account_request_add_cb(self, name, alias):
        def add_buddy_cb():
            self.__is_service.add_buddy(name, alias)

        if alias:
            message = "%s (%s)<br>added you.<br>Add buddy?" % (name, alias)
        else:
            message = "%s<br>Added you.<br>Add buddy?" % (name)

        msg = MessageDlgController(confirm_cb=add_buddy_cb)
        msg.show_message(message, buttons=2, title="BUDDY REQUEST")

    def _account_request_authorize_cb(self, name, alias, on_list):
        def authorize_cb():
            self.__is_service.authorize_buddy(name, True)
            if not on_list:
                self.__is_service.add_buddy(name, alias)

        def deny_cb():
            self.__is_service.authorize_buddy(name, False)
            if on_list:
                self.__is_service.remove_buddy(name)

        if alias:
            message = "%s (%s)<br>wants to add you.<br>" \
                    "Authorize and add buddy?" % (name, alias)
        else:
            message = "%s<br>wants to add you.<br>" \
                    "Authorize and add buddy?" % (name)

        msg = MessageDlgController(confirm_cb=authorize_cb, cancel_cb=deny_cb)
        msg.show_message(message, buttons=2, title="BUDDY REQUEST")

    def _received_msg_cb(self, username, alias, stripped):
        greeting_pattern = re.compile(GREETING_PATTERN, re.MULTILINE)
        accepted_pattern = re.compile(ACCEPTED_PATTERN, re.MULTILINE)
        rejected_pattern = re.compile(REJECTED_PATTERN, re.MULTILINE)
        close_conn_pattern = re.compile(CLOSE_CONN_PATTERN, re.MULTILINE)
        data_pattern = re.compile(DATA_PATTERN, re.MULTILINE)

        key = username.split("/")[0].encode("utf-8")
        alias = alias.encode("utf-8")

        if data_pattern.search(stripped):
            return self.buddy_model.received_data_msg(key, stripped)
        elif greeting_pattern.search(stripped):
            return self._received_greeting_msg(key, alias)
        elif accepted_pattern.search(stripped):
            return self._received_accepted_msg(key, alias)
        elif rejected_pattern.search(stripped):
            return self._received_rejected_msg(key)
        elif close_conn_pattern.search(stripped):
            return self.buddy_model.received_close_conn_msg(key)
        else:
            # Someone not using Carman sent a message to you. Send a
            # human-readable message to this buddy.
            self.__is_service.send_message(username, NO_IM_MESSAGE % CLOSE_CONN_FMT)
            return False

    def _received_greeting_msg(self, name, alias):
        def user_accepted_cb():
            self.send_accept_conn(name)
            self.buddy_model.connect_buddy(name)

        def user_rejected_cb():
            self.send_reject_conn(name)

        if self.buddy_model.has_request(name):
            # Request already in progress
            pass
        elif self.buddy_model.is_buddy_connected(name):
            # Buddy already connected
            pass
        elif self.buddy_model.size >= BuddyModel.MAX_BUDDIES:
            # Reached maximum number of connected buddies
            self.send_reject_conn(name)
            pass
        else:
            self.buddy_model.add_received_request_timeout(name, alias, \
                user_accepted_cb, user_rejected_cb)
            return True

    def _received_accepted_msg(self, name, alias):
        if self.buddy_model.has_request(name):
            self.buddy_model.remove_request(name, accepted=True)

            if not self.buddy_model.is_buddy_connected(name):
                self.buddy_model.connect_buddy(name)
        return True

    def _received_rejected_msg(self, name):
        if self.buddy_model.has_request(name):
            self.buddy_model.remove_request(name)

            msg = MessageDlgController()
            msg.show_message("%s<br> rejected connection request" % name)
        return True

    def add_buddy_signed_on_cb(self, cb):
        if callable(cb) and cb not in self.__buddy_signed_on_cbs:
            self.__buddy_signed_on_cbs.append(cb)

    def del_buddy_signed_on_cb(self, cb):
        if cb in self.__buddy_signed_on_cbs:
            self.__buddy_signed_on_cbs.remove(cb)

    def add_buddy_signed_off_cb(self, cb):
        if callable(cb) and cb not in self.__buddy_signed_off_cbs:
            self.__buddy_signed_off_cbs.append(cb)

    def del_buddy_signed_off_cb(self, cb):
        if cb in self.__buddy_signed_off_cbs:
            self.__buddy_signed_off_cbs.remove(cb)

    def add_connection_error_cb(self, cb):
        if callable(cb) and cb not in self.__connection_error_cbs:
            self.__connection_error_cbs.append(cb)

    def del_connection_error_cb(self, cb):
        if cb in self.__connection_error_cbs:
            self.__connection_error_cbs.remove(cb)

    def add_signed_on_cb(self, cb):
        if callable(cb) and cb not in self.__signed_on_cbs:
            self.__signed_on_cbs.append(cb)

    def del_signed_on_cb(self, cb):
        if cb in self.__signed_on_cbs:
            self.__signed_on_cbs.remove(cb)

    def add_signed_off_cb(self, cb):
        if callable(cb) and cb not in self.__signed_off_cbs:
            self.__signed_off_cbs.append(cb)

    def del_signed_off_cb(self, cb):
        if cb in self.__signed_off_cbs:
            self.__signed_off_cbs.remove(cb)

    def add_network_error_cb(self, cb):
        if callable(cb) and cb not in self.__network_error_cbs:
            self.__network_error_cbs.append(cb)

    def del_network_error_cb(self, cb):
        if cb in self.__network_error_cbs:
            self.__network_error_cbs.remove(cb)

    def get_buddy_alias(self, name):
        if self.__is_service:
            return self.__is_service.get_buddy_alias(name)
        else:
            return None

    def send_data_cb(self):
        mode = self.carmand_model.GetMode()

        if self.carmand_model.GetGPSDevice() == "none" and mode == "normal":
            gps_status = GPSModel.DISABLED
        else:
            gps_status = self.gps_model.Status()

        if self.carmand_model.GetOBDDevice() == "none" and mode == "normal":
            obd_status = OBDModel.DISABLED
        else:
            obd_status = self.obd_model.Status()

        message = DATA_FMT % (self.data[0], self.data[1], self.data[2], \
            self.data[3], self.data[4], self.data[5], gps_status, obd_status)

        for buddy in self.buddy_model.get_connected_buddies():
            self.__is_service.send_message(buddy, message)

        return True

    def send_request_authorization(self, name):
        if self.buddy_model.has_request(name):
            msg = MessageDlgController()
            msg.show_message("Connection request failed:<br>" \
                "Request already in progress", title="BUDDY REQUEST")
        elif self.buddy_model.is_buddy_connected(name):
            msg = MessageDlgController()
            msg.show_message("Connection request failed:<br>" \
                "Buddy already connected", title="BUDDY REQUEST")
        else:
            self.__is_service.send_message(name, \
                GREETING_MESSAGE % GREETING_FMT)
            self.buddy_model.add_sent_request_timeout(name)

    def send_accept_conn(self, name):
        self.__is_service.send_message(name, ACCEPTED_FMT)

    def send_reject_conn(self, name):
        self.__is_service.send_message(name, REJECTED_FMT)

    def send_close_connection(self, name):
        self.__is_service.send_message(name, CLOSE_CONN_FMT)
